@@ -75,8 +75,17 @@ class WSTerminalConnection(tornado.websocket.WebSocketClientConnection):
             if packet:
                 self._queue.put_nowait(packet.message)
 
+    def on_connection_close(self):
+        self._closed = True
+        if self._handler:
+            self._handler.on_connection_close()
+
     async def polling_packet_task(self):
         while not self._closed:
+            if self._queue.empty():
+                # Ensure the task exit in time
+                await asyncio.sleep(0.005)
+                continue
             packet = await self._queue.get()
             if packet["type"] == proto.EnumPacketType.REQUEST:
                 await self.handle_request(packet)
@@ -139,9 +148,12 @@ class WSTerminalClient(object):
         self._conn = WSTerminalConnection(url, headers, timeout, handler=self)
         self._loop = loop or asyncio.get_event_loop()
         self._workspace = None
-        self._shell_stdin = None
         self._running = True
 
+    def on_connection_close(self):
+        utils.logger.warn("[%s] Websocket connection closed" % self.__class__.__name__)
+        self.on_shell_exit()
+
     async def connect(self):
         return await self._conn.wait_for_connecting()
 
@@ -168,8 +180,7 @@ class WSTerminalClient(object):
                 await self.remove_directory(path)
             else:
                 await self.update_workspace(
-                    dir_tree["dirs"][name],
-                    path,
+                    dir_tree["dirs"][name], path,
                 )
         for name in dir_tree.get("files", {}):
             path = (root + "/" + name) if root else name
@@ -199,6 +210,14 @@ class WSTerminalClient(object):
         utils.logger.info("[%s] Remove file %s" % (self.__class__.__name__, file_path))
         await self._conn.send_request(proto.EnumCommand.REMOVE_FILE, path=file_path)
 
+    async def move_item(self, src_path, dst_path):
+        utils.logger.info(
+            "[%s] Move item %s to %s" % (self.__class__.__name__, src_path, dst_path)
+        )
+        await self._conn.send_request(
+            proto.EnumCommand.MOVE_ITEM, src_path=src_path, dst_path=dst_path
+        )
+
     async def on_directory_created(self, path):
         await self.create_directory(path)
 
@@ -214,6 +233,9 @@ class WSTerminalClient(object):
     async def on_file_removed(self, path):
         await self.remove_file(path)
 
+    async def on_item_moved(self, src_path, dst_path):
+        await self.move_item(src_path, dst_path)
+
     async def sync_workspace(self, workspace_path):
         if not os.path.isdir(workspace_path):
             raise RuntimeError("Workspace %s not exist" % workspace_path)
@@ -222,8 +244,7 @@ class WSTerminalClient(object):
             "%s%s" % (socket.gethostname(), workspace_path)
         )
         request = await self._conn.send_request(
-            proto.EnumCommand.SYNC_WORKSPACE,
-            workspace=workspace_hash,
+            proto.EnumCommand.SYNC_WORKSPACE, workspace=workspace_hash,
         )
         response = await self._conn.read_response(request)
         diff_result = self._workspace.make_diff(response["data"])
@@ -233,42 +254,49 @@ class WSTerminalClient(object):
         asyncio.ensure_future(self._workspace.watch())
 
     async def write_shell_stdin(self, buffer):
-        await self._conn.send_request(proto.EnumCommand.WRITE_STDIN, buffer=buffer)
+        try:
+            await self._conn.send_request(proto.EnumCommand.WRITE_STDIN, buffer=buffer)
+        except tornado.websocket.WebSocketClosedError:
+            utils.logger.error(
+                "[%s] Websocket connection lost" % self.__class__.__name__
+            )
+            self.on_connection_close()
 
     def on_shell_exit(self):
         self._running = False
-        if self._shell_stdin:
-            self._loop.remove_reader(self._shell_stdin)
-            self._shell_stdin = None
 
         async def exit_loop():
+            await asyncio.sleep(0.01)
             self._loop.stop()
 
         asyncio.ensure_future(exit_loop())
 
     async def create_shell(self, size):
         request = await self._conn.send_request(
-            proto.EnumCommand.CREATE_SHELL,
-            size=size,
+            proto.EnumCommand.CREATE_SHELL, size=size,
         )
         response = await self._conn.read_response(request)
         if response["code"]:
             raise RuntimeError("Create shell failed: %s" % response["message"])
         server_platform = response["platform"]
         if sys.platform != "win32":
-            self._shell_stdin = utils.UnixStdIn()
+            with utils.UnixStdIn() as shell_stdin:
 
-            def on_input():
-                char = self._shell_stdin.read(1)
-                if char == b"\n":
-                    char = b"\r"
-                if server_platform == "win32":
-                    if char == b"\x1b":
-                        chars = self._shell_stdin.read(2)
-                        char += chars  # Must send together
-                utils.safe_ensure_future(self.write_shell_stdin(char))
+                def on_input():
+                    char = shell_stdin.read(1)
+                    if char == b"\n":
+                        char = b"\r"
+                    if server_platform == "win32":
+                        if char == b"\x1b":
+                            chars = shell_stdin.read(2)
+                            char += chars  # Must send together
+                    utils.safe_ensure_future(self.write_shell_stdin(char))
 
-            self._loop.add_reader(self._shell_stdin, on_input)
+                self._loop.add_reader(shell_stdin, on_input)
+
+                while self._running:
+                    await asyncio.sleep(0.005)
+                self._loop.remove_reader(shell_stdin)
         else:
             import msvcrt
 
